/**
 * Copyright (C) 2017 Marcus Wittig
 *
 * This file is part of node-red-contrib-pimatic.
 *
 * node-red-contrib-pimatic is free software: you can redistribute it
 * and/or modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation, either version 2 of
 * the License, or(at your option) any later version.
 *
 * node-red-contrib-pimatic is distributed in the hope that it will
 * be useful, but WITHOUT ANY WARRANTY; without even the implied
 * warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program. If not, see <http://www.gnu.org/licenses/>.
 */
module.exports = function(RED) {
  "use strict";

  var util = require('util');
  var io = require('socket.io-client');
  var uuidv4 = require('uuid/v4');
  var _ = require('./helper');
  //var packageFile = require('../package.json');
  //
  // Pimatic Controller Node
  //

  function pimaticControllerNode(config) {
    RED.nodes.createNode(this, config);
    var node = this;
    // _.debugInit(node);
    node.pimaticVariables = {};
    node.pimaticSubscribedVariables = {};
    node.pimaticSocket = null;
    node.log('Started' + JSON.stringify(config));

    function addVariableHandler(variable) {
      var value = {
        name: variable.name,
        unit: variable.unit,
        type: variable.type,
        value: variable.value,
        readonly: variable.readonly,
        time: Date.now()
      };

      node.pimaticVariables[variable.name] = value;

      // check subscriptions - notify subscribers the new value
      if (node.pimaticSubscribedVariables.hasOwnProperty(variable.name)) {
        var subscribers = node.pimaticSubscribedVariables[variable.name];
        for (var x = 0; x < subscribers.length; ++x) {
          subscribers[x].emit('pimatic-variable-value-changed', value)
        }
      }
    }

    function removeVariableHandler(variable) {
      node.log('variable removed: ' + variable.name);
      if (node.pimaticVariables.hasOwnProperty(variable.name)) {
        delete node.pimaticVariables[variable.name];
      }

      // check subscriptions - notify subscribers the variable no longer exists
      if (node.pimaticSubscribedVariables.hasOwnProperty(variable.name)) {
        var subscribers = node.pimaticSubscribedVariables[variable.name];
        for (var x = 0; x < subscribers.length; ++x) {
          subscribers[x].emit('pimatic-variable-not-found', variable.name)
        }
      }
    }

    function stopWebSocket() {
      if (node.pimaticSocket != null) {
        node.pimaticSocket.close();
        node.pimaticSocket = null;
        node.log('closed web socket');
      }
    }

    function startWebSocket() {
      function socketErrorHandler(errorMessage) {
        if (errorMessage.hasOwnProperty('message')) {
          errorMessage = 'connection error: ' + errorMessage.message
        }
        else {
          errorMessage = 'connection error: ' + errorMessage;
        }
        node.error(errorMessage);
        for (var key in node.pimaticSubscribedVariables) {
          if (node.pimaticSubscribedVariables.hasOwnProperty(key)) {
            var subscribers = node.pimaticSubscribedVariables[key];
            for (var x = 0; x < subscribers.length; ++x) {
              subscribers[x].emit('pimatic-connection-error', errorMessage);
            }
          }
        }
      }

      if (node.pimaticSocket == null) {
        node.pimaticSocket = io(_.getConnectorURL(config), {
          reconnection: true,
          reconnectionDelay: 1000,
          reconnectionDelayMax: 3000,
          timeout: 20000,
          forceNew: true
        });
        node.log('first subscriber - starting web socket');

        node.pimaticSocket.on('error', socketErrorHandler);
        node.pimaticSocket.on('connect_error', socketErrorHandler);
        node.pimaticSocket.on('variableAdded', addVariableHandler);
        node.pimaticSocket.on('variableChanged', addVariableHandler);
        node.pimaticSocket.on('variableRemoved', removeVariableHandler);

        node.pimaticSocket.on('variableValueChanged', function(varEvent) {
          var variable = {
            name: varEvent.variableName,
            value: varEvent.variableValue
          };
          //node.log('variableValueChanged event: ' + variable.name);
          if (node.pimaticVariables.hasOwnProperty(variable.name)) {
            // augment type info as variableValueChanged event does not provide for it
            variable = Object.assign({}, node.pimaticVariables[variable.name], variable);
          }
          else {
            variable = Object.assign(variable, {
              type: 'unknown',
              unit: '',
              readonly: true
            });
          }
          node.pimaticVariables[variable.name] = variable;

          if (node.pimaticSubscribedVariables.hasOwnProperty(variable.name)) {
            var subscribers = node.pimaticSubscribedVariables[variable.name];
            for (var x = 0; x < subscribers.length; ++x) {
              subscribers[x].emit('pimatic-variable-value-changed', variable);
            }
          }
        });

        node.pimaticSocket.on('variables', function(variables) {
          node.log('variables event received variables: ' + variables.length);
          for (var x = 0; x < variables.length; ++x) {
            addVariableHandler(variables[x]);
          }
        });
      }
    }

    node.registerVariable = function (variableName, subscriber) {
      startWebSocket();

      if (node.pimaticSubscribedVariables.hasOwnProperty(variableName)) {
        node.pimaticSubscribedVariables[variableName].push(subscriber)
      }
      else {
        node.pimaticSubscribedVariables[variableName] = [subscriber];
      }

      setTimeout(function () {
        if (node.pimaticVariables.hasOwnProperty(variableName)) {
          var variable = node.pimaticVariables[variableName];
          subscriber.emit('pimatic-variable-value-changed', variable);
        }
        else {
          // variable not found
          subscriber.emit('pimatic-variable-not-found', variableName)
        }
      }, 1000);
    };

    node.invokeAction = function(action, params) {
      var id = uuidv4();

      return new Promise(function (resolve, reject) {

        if (node.pimaticSocket == null) {
          node.error(util.format("invokeAction (%s) failed: %s", action, 'socket not ready'));
          return reject('socket not ready')
        }

        var callResultHandler = function (callResult) {
          if (callResult.id == id) {
            node.pimaticSocket.removeListener('error', socketErrorHandler);
            node.pimaticSocket.removeListener('connect_error', socketErrorHandler);
            node.pimaticSocket.removeListener('callResult', callResultHandler);
            if (callResult.success) {
              resolve(callResult.result)
            }
            else {
              node.error(util.format("invokeAction (%s) failed: %s", action, '' + callResult.error));
              reject(callResult.error)
            }
          }
        };

        var socketErrorHandler = function (errorMessage) {
          node.pimaticSocket.removeListener('callResult', callResultHandler);
          if (errorMessage.hasOwnProperty('message')) {
            errorMessage = 'connection error: ' + errorMessage.message
          }
          else {
            errorMessage = 'connection error: ' + errorMessage;
          }
          node.error(util.format("invokeAction (%s) failed: %s", action, errorMessage));
          reject(errorMessage);
        };

        node.pimaticSocket.emit('call', {
          id: id,
          action: action,
          params: params
        });

        node.pimaticSocket.on('callResult', callResultHandler);
        node.pimaticSocket.once('error', socketErrorHandler);
        node.pimaticSocket.once('connect_error', socketErrorHandler);
      });
    };

    node.setVariableValue = function(name, value, unit) {
      var params = {
        name: name,
        valueOrExpression: value,
        type: 'value',
        unit: unit
      };
      return node.invokeAction('updateVariable', params).then(function(callResult) {
        return Promise.resolve(callResult.variable)
      })
    };

    node.getVariableValue = function(name) {
      var params = {
        name: name
      };
      return node.invokeAction('getVariableByName', params).then(function(callResult) {
        return Promise.resolve(callResult.variable)
      })
    };

    node.executeDeviceAction = function(deviceId, actionName, actionParams) {
      actionParams = actionParams || {};
      var params = Object.assign({
        deviceId: deviceId,
        actionName: actionName
      }, actionParams);
      return node.invokeAction('callDeviceAction', params)
    };

    node.deregisterVariable = function (variableName, subscriber) {
      if (node.pimaticSubscribedVariables.hasOwnProperty(variableName)) {
        var subscribers = node.pimaticSubscribedVariables[variableName];
        var x = subscribers.length;
        while (x--) {
          if (subscribers[x] == subscriber) {
            node.log('Delete subscriber for variable: ' + variableName);
            subscribers.splice(x, 1);
          }
        }
        if (subscribers.length === 0) {
          node.log('No more subscribers for variable: ' + variableName);
          delete node.pimaticSubscribedVariables[variableName]
        }
        node.log('Remaining subscribed vars: ' + Object.keys(node.pimaticSubscribedVariables));
        if (Object.keys(node.pimaticSubscribedVariables).length === 0) {
          node.log('no subscribers - stopping web socket');
          stopWebSocket();
        }
      }
    };

    node.on('close', function(done) {
      node.log('closing pimaticControllerNode');
      stopWebSocket();
      node.status({});
      node.log('closed pimaticControllerNode');
      done();
    });
  }

  //
  // Pimatic Variable Input
  //

  function pimaticVariableInputNode(config) {
    RED.nodes.createNode(this, config);
    var node = this;
    // _.debugInit(node);

    if (_.hasStringValue(config.variable, true) &&
      _.hasStringValue(config.controller, true)) {
      node.name = config.name || '(' + config.variable + ')';
      var pimaticController = RED.nodes.getNode(config.controller);
      node.log(JSON.stringify(config));

      node.on('pimatic-variable-value-changed', function(varEvent) {
        node.status({fill: 'green', shape: 'ring', text: 'ok'});

        if (varEvent.name == config.variable) {
          node.log('pimatic-variable-value-changed for ' + config.variable)
          var context = node.context();
          if (context.get('value') != varEvent.name || !config.filter) {
            context.set('value', varEvent.value);
            var msg = Object.assign({
              payload: varEvent.value
            }, varEvent);
            delete msg.value;
            node.send(msg);
          }
        }
      });

      node.on('input', function(msg) {
        // query variable
        // TODO: Should be replaced with a pimatic query operation

        if (pimaticController.pimaticVariables.hasOwnProperty(config.variable)) {
          var variable = pimaticController.pimaticVariables[config.variable];
          var context = node.context();
          if (context.get('value') != variable.variableValue|| !config.filter) {
            context.set('value', variable.variableValue);
            var msg = Object.assign({
              payload: variable.value
            }, variable);
            delete msg.value;
            node.send(msg);
          }
        }
      });

      node.on('pimatic-variable-not-found', function(variableName) {
        node.error('variable not found: ' + variableName);
        node.status({fill: 'red', shape: 'ring', text: 'variable not found'});
      });

      node.on('pimatic-connection-error', function(errorMessage) {
        node.error(errorMessage);
        node.status({fill: 'red', shape: 'ring', text: '' + errorMessage});
      });

      node.on('close', function(done) {
        node.log('closing pimaticVariableInputNode');
        pimaticController.deregisterVariable(config.variable, node);
        node.removeAllListeners('input');
        node.removeAllListeners('pimatic-variable-value-changed');
        node.removeAllListeners('pimatic-variable-not-found');
        node.removeAllListeners('pimatic-connection-error');
        node.status({});
        node.log('closed pimaticVariableInputNode');
        done();
      });

      pimaticController.registerVariable(config.variable, node);
    }
    else {
      node.error('invalid node configuration');
      node.status({fill: 'red', shape: 'ring', text: 'invalid node configuration'});
    }
  }

  //
  // Pimatic Variable Output
  //

  function pimaticVariableOutputNode(config) {
    RED.nodes.createNode(this, config);
    var node = this;
    // _.debugInit(node);

    if (_.hasStringValue(config.variable, true) &&
      _.hasStringValue(config.controller, true)) {

      node.name = config.name || '(' + config.variable + ')';
      var pimaticController = RED.nodes.getNode(config.controller);
      node.log(JSON.stringify(config));

      node.on('pimatic-variable-value-changed', function(varEvent) {
        if (varEvent.name == config.variable) {
          node.log('pimatic-variable-value-changed for ' + config.variable);

          if (varEvent.readonly) {
            node.status({fill: 'red', shape: 'ring', text: 'variable is readonly'});
          }
          else {
            node.status({fill: 'green', shape: 'ring', text: 'ok'});
          }
        }
      });

      node.on('input', function(msg) {
        node.log("input");

        if (pimaticController.pimaticVariables.hasOwnProperty(config.variable)
          && !pimaticController.pimaticVariables[config.variable].readonly) {
          node.log("setVariable");
          pimaticController.setVariableValue(config.variable, config.value || msg.payload, config.unit)
            .then(function(variable) {
              pimaticController.pimaticVariables[config.variable] = variable;
              var context = node.context();
              context.set('value', variable.variableValue);

              var msg = Object.assign({}, variable, {
                payload: variable.value,
                time: Date.now()
              });
              delete msg.value;
              node.send(msg);
            })
            .catch(function(error) {
              node.error(util.format('setVariable (%s) failed: %s', config.variable, '' + error))
              node.status({fill: 'red', shape: 'ring', text: '' + error});
            });
        }
      });

      node.on('pimatic-variable-not-found', function(variableName) {
        node.error('variable not found: ' + variableName);
        node.status({fill: 'red', shape: 'ring', text: 'variable not found'});
      });

      node.on('pimatic-connection-error', function(errorMessage) {
        node.error(errorMessage);
        node.status({fill: 'red', shape: 'ring', text: '' + errorMessage});
      });

      node.on('close', function(done) {
        node.log('closing pimaticVariableInputNode');
        pimaticController.deregisterVariable(config.variable, node);
        node.removeAllListeners('input');
        node.removeAllListeners('pimatic-variable-value-changed');
        node.removeAllListeners('pimatic-variable-not-found');
        node.removeAllListeners('pimatic-connection-error');
        node.status({});
        node.log('closed pimaticVariableInputNode');
        done();
      });

      pimaticController.registerVariable(config.variable, node);
    }
    else {
      node.error('invalid node configuration');
      node.status({fill: 'red', shape: 'ring', text: 'invalid node configuration'});
    }
  }

  //
  // Pimatic Device Action Node
  //

  function pimaticDeviceActionNode(config) {
    RED.nodes.createNode(this, config);
    var node = this;
    // _.debugInit(node);

    if (_.hasStringValue(config.deviceId, true) &&
      _.hasStringValue(config.action, true) &&
      _.hasStringValue(config.controller, true)) {

      node.name = config.name || util.format('(%s.%s)', config.deviceId, config.action);
      var pimaticController = RED.nodes.getNode(config.controller);
      node.log(JSON.stringify(config));

      node.on('input', function(msg) {
        node.log("input");

        if (pimaticController.pimaticVariables.hasOwnProperty(config.variable)
          && !pimaticController.pimaticVariables[config.variable].readonly) {
          node.log("setVariable");
          pimaticController.setVariableValue(config.variable, config.value || msg.payload, config.unit)
            .then(function(variable) {
              pimaticController.pimaticVariables[config.variable] = variable;
              var context = node.context();
              context.set('value', variable.variableValue);

              var msg = Object.assign({}, variable, {
                payload: variable.value,
                time: Date.now()
              });
              delete msg.value;
              node.send(msg);
            })
            .catch(function(error) {
              node.error(util.format('setVariable (%s) failed: %s', config.variable, '' + error))
              node.status({fill: 'red', shape: 'ring', text: '' + error});
            });
        }
      });

      node.on('pimatic-connection-error', function(errorMessage) {
        node.error(errorMessage);
        node.status({fill: 'red', shape: 'ring', text: '' + errorMessage});
      });

      node.on('close', function(done) {
        node.log('closing pimaticVariableInputNode');
        //pimaticController.deregisterVariable(config.variable, node);
        node.removeAllListeners('input');
        node.removeAllListeners('pimatic-variable-value-changed');
        node.removeAllListeners('pimatic-variable-not-found');
        node.removeAllListeners('pimatic-connection-error');
        node.status({});
        node.log('closed pimaticVariableInputNode');
        done();
      });

      //pimaticController.registerVariable(config.variable, node);
    }
    else {
      node.error('invalid node configuration');
      node.status({fill: 'red', shape: 'ring', text: 'invalid node configuration'});
    }
  }


  //
  // Register Node Types
  //
  RED.nodes.registerType('controller', pimaticControllerNode);
  RED.nodes.registerType('get variable', pimaticVariableInputNode);
  //RED.nodes.registerType('set variable', pimaticVariableOutputNode);
  //RED.nodes.registerType('device action', pimaticDeviceActionNode);
  //RED.log.info(util.format('%s@%s started', packageFile.name, packageFile.version));
};
/**
 * Created by marcus on 14.07.2017.
 */
